%{
/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: lexer.l 284 2017-10-08 08:00:31Z cau $
 */

#include "system.h"
char str[MAXTOKENSIZE];
char *plstr=str;
char *pmaxstr = str+MAXTOKENSIZE-1;
int yycolnum = 1;
int64_t cint;
char Linetext[MAXLINESIZE];
char *plinetext=Linetext;
char *plinemax=Linetext+MAXLINESIZE-1;
char *pyytext;
#define YY_USER_ACTION {yylloc.first_line = yylineno; \
        yylloc.first_column = yycolnum;                 \
        yycolnum=yycolnum+yyleng;                         \
        yylloc.last_column=yycolnum-1;    \
        yylloc.last_line = yylineno;\
        pyytext= yytext;\
        while ((*plinetext++ = *yytext++) != '\0') { \
         if (plinetext > plinemax) \
               syntax_error (IPTL); \
        }\
        *(plinetext < plinemax ? plinetext : plinemax) = '\0'; \
        *plinetext--;\
        yytext=pyytext;\
       }

%}
%option always-interactive
%option noyywrap
%option yylineno
%option bison-bridge
%s C S SS FF

%%

\n			{
                         yycolnum = 1;
                         Linetext[0]='\0';plinetext=Linetext;
                        }
<C>[^*\n]*		;
<C>"*"+[^*/\n]*         ;
<C>"*"+"/"		{
			BEGIN 0;/* finish reading comment */
			}
<C><<EOF>>              {
                         syntax_error (IDN);
                         EOFILE=1;
                         BEGIN 0;
                         return(END);
                        }
<S>\\n   		{ 
                        *(plstr < pmaxstr ? plstr++ : pmaxstr) = '\n';
                        if (plstr > pmaxstr)
                          syntax_error (STL);
   			}
<S>\\r   		{ 
                        *(plstr < pmaxstr ? plstr++ : pmaxstr) = '\r';
                        if (plstr > pmaxstr)
                          syntax_error (STL);
			}
<S>\\t   		{ 
                        *(plstr < pmaxstr ? plstr++ : pmaxstr) = '\t';
                        if (plstr > pmaxstr)
                          syntax_error (STL);
			}
<S>\\f   		{ 
                        *(plstr < pmaxstr ? plstr++ : pmaxstr) = '\f';
                        if (plstr > pmaxstr)
                          syntax_error (STL);
			}
<S>\\b   		{ 
                        *(plstr < pmaxstr ? plstr++ : pmaxstr) = '\b';
                        if (plstr > pmaxstr)
                          syntax_error (STL);
			}
<S>\\\\                 {
                        *(plstr < pmaxstr ? plstr++ : pmaxstr) = '\\';
                        if (plstr > pmaxstr)
                          syntax_error (STL);
                        }
<S>\\[\"]               {
                        *(plstr < pmaxstr ? plstr++ : pmaxstr) = '\"';
                         if (plstr > pmaxstr)
                          syntax_error (STL);
                        }
<S>[^\\"\n]*		{
			while ((*plstr++ = *yytext++) != '\0') {
                         if (plstr > pmaxstr)
                          syntax_error (STL);
                        }
                        *plstr-- = *yytext--;
			}
<S>\"			{
			BEGIN 0;		
                        /* Finish reading string. */
                        *(plstr < pmaxstr ? plstr : pmaxstr) = '\0';
			yylval->nodetype = string_node(strsave(str));
			return(STRING);
			}
<S><<EOF>>              {
                        syntax_error (MQS);
                        EOFILE=1;
                         BEGIN 0;
                         return(END);
                        }
\"			{ 
			BEGIN S; 
                        /* Start to read a string. */
                        str[0] = '\0';plstr=str;  
			}
<SS>[^\'\n]*		{
			yymore();
			}
<SS>\'			{
			BEGIN 0;		/* Finish reading string. */
			yytext[yyleng-1] = '\0';/* Delete final "'". */
			yylval->nodetype = string_node(strsave(yytext));
			return(STRING);
			}
<SS><<EOF>>             {
                        syntax_error (MQS);
                         EOFILE=1;
                         BEGIN 0;
                         return(END);
                        }
\'                      {
                        BEGIN SS;/* Start to read a string. */
                        }
<FF>("+"|"-")?([0-9]*(".")?[0-9]+)((e|E)("+"|"-")?[0-9]+)?"$" {
                        BEGIN 0;		/* Finish reading float. */
			yytext[yyleng-1] = '\0';/* Delete final `$'. */
			yylval->nodetype = float_node(atof(yytext));
			return(FLOAT);
                        }
<FF>[^$\n]*             {
                        syntax_error (EIF);
                        EOFILE=1;
                         BEGIN 0;
                         return(END);
                        }
<FF><<EOF>>             {
                        syntax_error (MDF);
                        EOFILE=1;
                         BEGIN 0;
                         return(END);
                        }
"$"                     {
                        BEGIN FF;/* Start to read a float. */
                        }
[ \t\r]			;
"."                     {
                         return(END);
                        }
"+"                     {
                         return(PLUS);
                        }
"-"                     {
                         return(MINUS);
                        }
"*"                     {
                         return(MULT);
                        }
"|"                     {
                         return(BAR);
                        }
"/"                     {
                         return(DIV);
                        }
"~="                    {
                         return(NEQ);
                        }
"~"                     {
                         return(NOT);
                        }
".."                    {
                         return(TO);
                        }
"<-"                    {
                         return(ASSN);
                        }
":="                    {
                         return(UASSN);
                        }
":-"                    {
                         return(SKIPUASSN);
                        }
"<="                    {
                         return(LE);
                        }
">="                    {
                         return(GE);
                        }
"**"                    {
                         return(POWER);
                        }
"["                     {
                         return(LBRACK);
                        }
"]"                     {
                          return(RBRACK);
                        }
"="                     {
                          return(EQ);
                        }
"<"                     {
                          return(LT);
                        }
">"                     {
                          return(GT);
                        }
":"                     {
                          return(DO);
                        }
";"                     {
                          return(SEMI);
                        }
"("                     {
                          return(LPAREN);
                        }
")"                     {
                          return(RPAREN);
                        }
","                     { 
                          return(COMMA);
                        }
"{"                     {
                          return(LBRACE);
                        }
"}"                     {
                          return(RBRACE);
                        }
"#"                     {
                          return(HISTORY);
                        }
"existsf"               {
                          return(EXISTSF);
                        }
"@"                     {
                          return(REFER);
                        }
"?"                     {
                          return(HELP);
                        }
"!"                     {
                          return(DEREF);
                        }
"do"                    {
                          return(DO);
                        }
"box"                   {
                          return(ALWAYS);
                        }
"sometime"              {
                          return(SOMETIME);
                        }
"jump"                  {
                         return(LEAP);
                        }
"shift"                 {
                         return(NEXT);
                        }
"frame"                 {
                         return(MEM);
                        }
[a-zA-Z][a-zA-Z0-9_]*	{
                        if ( (yylval->nodetype=lookup (yytext)) == NULL) {
                         yylval->nodetype = install(NAME,strsave(yytext));
                        } 
			return(node_type(yylval->nodetype));
			}
[0-9]+	                {
                        errno=0;
                        cint = strtoll(yytext,(char **)NULL,10);
                        if (errno == ERANGE)
                         syntax_error(ERA);
			yylval->nodetype = int_node(cint);
			return(INT);
			}
"/*"			{
			BEGIN C;/* start to read & discard comment */
			}
<<EOF>>                 {
                        EOFILE=1;yyterminate();
                        }
\x03                    {
                        }
.			{
                        syntax_error(UNK);
			}
%%

NODEPTR symballoc()           /* allocate a new symbol */
{
    if (SymbolTableTop >= SymbolTableMax)
        more_symbol_table (MAXSYMBOLTABLESIZE);
    return (SymbolTableTop++);
}

unsigned hash(s)                         /* form hash value for string s */
char *s;
{
    register unsigned hashval;

    for (hashval = 0; *s != '\0'; )
        hashval = (hashval << 1) ^ (*s++);
    return(hashval % HASHTABLESIZE);
}

NODEPTR lookup(s)
char *s;
{
    register NODEPTR sp;
    for (sp = HashTable[hash (s)]; sp != NULL; sp = node_next(sp))
        if (strcmp (s, node_sym_name(sp)) == 0)
            return (sp);
    return (NULL);
}

NODEPTR install(type,name)
int type;
char *name;
{
    register NODEPTR sp;
    unsigned hashval;
    sp = symballoc ();
    node_type(sp) = type;
    node_sym_name(sp) = name;
        /*printf("installing %s \n",name);*/
    hashval = hash (node_sym_name(sp));
    node_next(sp) = HashTable[hashval];
    HashTable[hashval] = sp;
    
    return (sp);
}


void install_keyword(type,name)
int type;
char *name;
{
    Keyword[type-MINTOKEN]=name;
    install (type, name);
}

void install_reduction(type,name1,name2,name3)
int type;
PFI name1,name2,name3;
{
    Reduces[type-MINTOKEN][0]=name1;
    Reduces[type-MINTOKEN][1]=name2;
    Reduces[type-MINTOKEN][2]=name3;
}

void install_reduction_w(type,name1,name2,name3)
int type;
char *name1,*name2,*name3;
{
    Reduc[type-MINTOKEN][0]=name1;
    Reduc[type-MINTOKEN][1]=name2;
    Reduc[type-MINTOKEN][2]=name3;
}

/*
uninstall(s)
char *s;
{
    SYMBOLPTR * spp;
    for (spp = &HashTable[hash (s)]; *spp != NULL; spp = &((*spp) -> next)) {
        if (strcmp (s, (*spp) -> name) == 0) {
            *spp = (*spp) -> next;
            return (T);
        }
    }
    return(F);
}
*/


